"""
A collection of Widgets and Listeners designed specifically for the ECS system.
"""

from bear_hug.bear_utilities import BearECSException, BearLayoutException, \
    copy_shape
from bear_hug.ecs import Entity
from bear_hug.event import BearEvent
from bear_hug.widgets import Layout


class ECSLayout(Layout):
    """
    A Layout of entities.
    
    It is controlled entirely by events. Although Layour methods like
    `add_child` and `move_child` are not overloaded, their use is discouraged.
    
    In all events, entity IDs, not the actual Entity objects, are passed. Event
    conventions are as following:
    
    `BearEvent(event_type='ecs_move', event_value=(entity, x, y))`.
    Announces that the widget of the entity in question should be moved
    to (x; y). Behaviour for widgets trying to leave map edges is left to
    be determined by ECSLayout or its children.
    
    `BearEvent(event_type='ecs_remove', event_value=entity)`.
    Announces that the widget of the entity in question should be removed from
    the ECSLayout. Does not cause or imply the destruction of entity object.
    
    `BearEvent(event_type='ecs_add', event_value=(entity, x, y))`.
    Announces that the widget of the entity in question should be added to the
    ECSLayout at (x;y). The emission of this event implies the existence of the
    entity and its widget.
    
    This widget also provides the collision detection for all widgets within it.
    If a widget attempts to move into the position occupied by some other
    widget, the movement is not blocked, but the 'ecs_collision' event(s) get
    emitted. The event follows the following convention:
    
    `BearEvent(event_type='ecs_collision', event_value=(widget_moved.id,
                                                    widget_collided_into.id))`
    If the widget enters the screen border (ie will go outside it the next time
    it moves, assuming it keeps the direction), the event takes the following
    form:
    `BearEvent(event_type='ecs_collision', event_value=(widget_moved.id,
                                                        None))`
    """
    
    def __init__(self, chars, colors):
        super().__init__(chars, colors)
        self.entities = {}
        self.widgets = {}
        self.need_redraw = False
    
    def add_entity(self, entity):
        """
        Register the entity to be displayed. Assumes that the entity has a
        widget already.
        
        The entity is not actually shown until the `'ecs_add'` event is emitted
        :return:
        """
        if not isinstance(entity, Entity):
            raise BearECSException('Cannot add non-Entity to ECSLayout')
        self.entities[entity.id] = entity
        self.widgets[entity.id] = entity.widget.widget
    
    def remove_entity(self, entity_id):
        """
        Forget about the registered entity and its widget.
        Does not imply or cause the destruction of Entity object or any of its
        Component objects (except if this was the last reference). Making sure
        that the entity is removed cleanly is someone else's job.
        :param entity_id:
        :return:
        """
        if entity_id not in self.entities:
            raise BearECSException('Attempting to remove nonexistent entity {} from ESCLayout'.
                                   format(entity_id))
        self.remove_child(self.entities[entity_id].widget.widget)
        del self.entities[entity_id]
        
    def on_event(self, event):
        # React to the events
        r = []
        if event.event_type == 'ecs_move':
            entity_id, x, y = event.event_value
            # Checking if collision events need to be emitted
            # Check for collisions with border
            if x < 0 or x+self.entities[entity_id].widget.size[0]\
                 > len(self.chars[0]) or y < 0 or \
                 y + self.entities[entity_id].widget.size[1] > len(self.chars):
                r.append(BearEvent(event_type='ecs_collision',
                                   event_value=(entity_id, None)))
            else:
                # Apparently no collision with a border, can safely move
                self.move_child(self.widgets[entity_id], (x, y))
                self.need_redraw = True
                collided = set()
                for y_offset in range(self.entities[entity_id].widget.size[1]):
                    for x_offset in range(self.entities[entity_id].widget.size[0]):
                        for other_widget in self._child_pointers[y+y_offset]\
                            [x+x_offset]:
                            # Child_pointers is ECS-agnostic and stores pointers
                            # to the actual widgets
                                collided.add(other_widget)
                collided_ent_ids = set()
                for child in self.entities:
                    if child != entity_id and \
                            self.entities[child].widget.widget in collided:
                        collided_ent_ids.add(child)
                for child in collided_ent_ids:
                    r.append(BearEvent('ecs_collision', (entity_id, child)))
        elif event.event_type == 'ecs_create':
            self.add_entity(event.event_value)
            self.need_redraw = True
        elif event.event_type == 'ecs_destroy':
            self.remove_entity(event.event_value)
            self.need_redraw = True
        elif event.event_type == 'ecs_remove':
            self.remove_child(self.entities[event.event_value].widget.widget)
            self.need_redraw = True
        elif event.event_type == 'ecs_add':
            entity_id, x, y = event.event_value
            self.add_child(self.widgets[entity_id], (x, y))
            self.need_redraw = True
        elif event.event_type == 'ecs_update':
            # Some widget has decided it's time to redraw itself
            self.need_redraw = True
        elif event.event_type == 'service' and event.event_value == 'tick_over'\
                and self.need_redraw:
            self._rebuild_self()
            self.terminal.update_widget(self)
            self.need_redraw = False
        if r:
            return r
        
    def __repr__(self):
        # It's unlikely that repr(layout) is ever gonna be necessary.
        # And it's very bad to try and serialize them. Widget serialization is
        # only intended for use with the Widgets that can be part of a
        # WidgetComponent. Layouts are sorta frontend for the entire ECS system,
        # not a part of it.
        raise BearECSException('ECSLayout and its children are not meant to be stored via repr')


class ScrollableECSLayout(Layout):
    """
    A Layout that can show only a part of its surface and supports ECS system.

    Like a ScrollableLayout, accepts `chars` and `colors` on creation, which
    should be the size of the entire layout, not just the visible area.
    The latter is initialized by `view_pos` and `view_size` arguments.
    
    Like an ECSLayout, supports all 'ecs_*' events. Unlike it, this class also
    supports two event types:
    
    `BearEvent(event_type='ecs_scroll_by', event_value=(x, y)). Shifts visible
    area by x chars horizontally and by y chars vertically.
    
    `BearEvent(event_type='ecs_scroll_to', event_value=(x, y). Moves visible
    area to (x, y).
    
    If incorrect value is provided to either of these events, it is silently
    ignored.
    """

    # This class is basically a copypaste of pieces from
    # widgets.ScrollableLayout and ecs_widgets.ECSLayout. These classes were not
    # merged or inherited from each other or something because they both can be
    # useful without each other's capabilities (and overhead). And multiple
    # inheritance is plain evil.
    def __init__(self, chars, colors,
                 view_pos=(0, 0), view_size=(10, 10)):
        super().__init__(chars, colors)
        if not 0 <= view_pos[0] <= self.width - view_size[0] \
                or not 0 <= view_pos[1] <= self.height - view_size[1]:
            raise BearLayoutException('Initial viewpoint outside ' +
                                      'ScrollableLayout')
        if not 0 < view_size[0] <= len(chars[0]) \
                or not 0 < view_size[1] <= len(chars):
            raise BearLayoutException('Invalid view field size')
        self.entities = {}
        self.widgets = {}
        self.view_pos = view_pos[:]
        self.view_size = view_size[:]
        self._rebuild_self()
    
    def _rebuild_self(self):
        """
        Same as `Layout()._rebuild_self`, but all child positions are also
        offset by `view_pos`. Obviously, only `view_size[1]` lines
        `view_size[0]` long are set as `chars` and `colors`.
        :return:
        """
        chars = [[' ' for x in range(self.view_size[0])] \
                 for y in range(self.view_size[1])]
        colors = copy_shape(chars, None)
        for line in range(self.view_size[1]):
            for char in range(self.view_size[0]):
                for child in self._child_pointers[self.view_pos[1] + line] \
                                     [self.view_pos[0] + char][::-1]:
                    # Addressing the correct child position
                    c = child.chars[
                        self.view_pos[1] + line - self.child_locations[child][
                            1]] \
                        [self.view_pos[0] + char - self.child_locations[child][
                            0]]
                    if c != ' ':
                        # Spacebars are used as empty space and are transparent
                        chars[line][char] = c
                        break
                colors[line][char] = \
                    child.colors[
                        self.view_pos[1] + line - self.child_locations[child][
                            1]] \
                        [self.view_pos[0] + char - self.child_locations[child][
                        0]]
        self.chars = chars
        self.colors = colors
    
    def resize_view(self, new_size):
        # TODO: support resizing view.
        # This will require updating the pointers in terminal or parent layout
        pass
    
    def scroll_to(self, pos):
        """
        Move field of view to `pos`.

        Raises `BearLayoutException` on incorrect position
        :param pos: tuple of ints
        :return:
        """
        if not (len(pos) == 2 and all((isinstance(x, int) for x in pos))):
            raise BearLayoutException('Field of view position should be 2 ints')
        if not 0 <= pos[0] <= len(self._child_pointers[0]) - self.view_size[0] \
                or not 0 <= pos[1] <= len(self._child_pointers) - \
                       self.view_size[1]:
            raise BearLayoutException('Scrolling to invalid position')
        self.view_pos = pos
    
    def scroll_by(self, shift):
        """
        Move field of view by `shift[0]` to the right and by `shift[1]` down.

        Raises `BearLayoutException` on incorrect position
        :param shift: tuple of ints
        :return:
        """
        pos = (self.view_pos[0] + shift[0], self.view_pos[1] + shift[1])
        self.scroll_to(pos)

    def add_entity(self, entity):
        """
        Register the entity to be displayed. Assumes that the entity has a
        widget already.

        The entity is not actually shown until the `'ecs_add'` event is emitted
        :return:
        """
        if not isinstance(entity, Entity):
            raise BearECSException('Cannot add non-Entity to ECSLayout')
        self.entities[entity.id] = entity
        self.widgets[entity.id] = entity.widget.widget

    def remove_entity(self, entity_id):
        """
        Forget about the registered entity and its widget.
        Does not imply or cause the destruction of Entity object or any of its
        Component objects (except if this was the last reference). Making sure
        that the entity is removed cleanly is someone else's job.
        :param entity_id:
        :return:
        """
        if entity_id not in self.entities:
            raise BearECSException(
                'Attempting to remove nonexistent entity {} from ESCLayout'.
                format(entity_id))
        self.remove_child(self.entities[entity_id].widget.widget)
        del self.entities[entity_id]

    def on_event(self, event):
        # React to the events
        r = []
        if event.event_type == 'ecs_move':
            entity_id, x, y = event.event_value
            # Checking if collision events need to be emitted
            # Check for collisions with border
            try:
                if x < 0 or x + self.entities[entity_id].widget.size[0] \
                        > len(self._child_pointers[0]) or y < 0 or \
                        y + self.entities[entity_id].widget.size[1] > len(
                        self._child_pointers):
                    r.append(BearEvent(event_type='ecs_collision',
                                       event_value=(entity_id, None)))
                else:
                    # Apparently no collision with a border, can safely move
                    self.move_child(self.widgets[entity_id], (x, y))
                    self.need_redraw = True
                    collided = set()
                    for y_offset in range(
                            self.entities[entity_id].widget.size[1]):
                        for x_offset in range(
                                self.entities[entity_id].widget.size[0]):
                            for other_widget in \
                            self._child_pointers[y + y_offset] \
                                    [x + x_offset]:
                                # Child_pointers is ECS-agnostic and stores pointers
                                # to the actual widgets
                                collided.add(other_widget)
                    # TODO: optimize to avoid checking all entities in collision detector
                    # Probably just storing all entity ids along with child pointers
                    collided_ent_ids = set()
                    for child in self.entities:
                        if child != entity_id and \
                                self.entities[child].widget.widget in collided:
                            collided_ent_ids.add(child)
                    for child in collided_ent_ids:
                        r.append(BearEvent('ecs_collision', (entity_id, child)))
            except KeyError:
                # In some weird cases 'ecs_move' events can be emitted after the
                # entity got destroyed
                return
        elif event.event_type == 'ecs_create':
            self.add_entity(event.event_value)
            self.need_redraw = True
        elif event.event_type == 'ecs_destroy':
            self.remove_entity(event.event_value)
            self.need_redraw = True
        elif event.event_type == 'ecs_remove':
            self.remove_child(
                self.entities[event.event_value].widget.widget)
            self.need_redraw = True
        elif event.event_type == 'ecs_add':
            entity_id, x, y = event.event_value
            self.add_child(self.widgets[entity_id], (x, y))
            self.need_redraw = True
        elif event.event_type == 'ecs_scroll_to':
            try:
                self.scroll_to(event.event_value)
                self.need_redraw = True
            except BearLayoutException:
                pass
        elif event.event_type == 'ecs_scroll_by':
            try:
                self.scroll_by(event.event_value)
                self.need_redraw = True
            except BearLayoutException:
                pass
        elif event.event_type == 'ecs_update':
            # Some widget has decided it's time to redraw itself
            self.need_redraw = True
        elif event.event_type == 'service' and event.event_value == 'tick_over' \
                and self.need_redraw:
            self._rebuild_self()
            self.terminal.update_widget(self)
            self.need_redraw = False
        
        if r:
            return r

    def __repr__(self):
        # It's unlikely that repr(layout) is ever gonna be necessary.
        # And it's very bad to try and serialize them. Widget serialization is
        # only intended for use with the Widgets that can be part of a
        # WidgetComponent. Layouts are sorta frontend for the entire ECS system,
        # not a part of it.
        raise BearECSException(
            'ECSLayout and its children are not meant to be stored via repr')



"""
A collection of random functions useful for bear_hug
"""

from copy import deepcopy


def shapes_equal(a, b):
    """
    Tests if two nested lists are of the same shape
    :param a: list
    :param b: list
    :return: bool
    """
    if len(a) != len(b):
        return False
    if any(isinstance(x, list) and isinstance(y, list) and len(x) != len(y)
           for x, y in zip(a, b)):
        return False
    return True


def copy_shape(l, value=None):
    """
    Takes a nested list and returns the new list of the same shape, completely
    filled with the same value.
    Works incorrectly with mutable types (particularly containers) because it
    fills the returned list with (pointers to) the same list, not independent
    copies and adding to either of them would affect all. Since this function
    gets called pretty often with primitives, eg None, and almost never with
    lists, it is left as is and callers are to be made sure their lists are OK.
    :param l: initial list
    :param value: value to fill the list with
    :return:
    """
    r = []
    for i in l:
        if isinstance(i, list):
            r.append(copy_shape(i, value))
        else:
            r.append(value)
    return r


def slice_nested(l, slice_pos, slice_size):
    """
    Slice the nested list
    :param l:
    :param slice_pos:
    :return:
    """
    r = []
    for y in range(slice_pos[1], slice_pos[1] + slice_size[1]):
        line = []
        for x in range(slice_pos[0], slice_pos[0] + slice_size[0]):
            line.append(l[y][x])
        r.append(line)
    return r


def rotate_list(l):
    """
    Take a nested list of (x, y) dimensions, return an (y, x) list
    :param l:
    :return:
    """
    # Without loss of generality, presume list is row-first and we need it
    # column-first
    r = [[None for x in range(len(l))] for x in range(len(l[0]))]
    for row_index, row in enumerate(l):
        for column_index, column_value in enumerate(row):
            r[column_index][row_index] = column_value
    return r
    

def rectangles_collide(pos1, size1, pos2, size2):
    """
    Return True if the rectangles collide
    Rectangles are supplied in [x,y], [xsize, ysize] form with the left corner
    and size. Assumes positions and sizes to be sorted
    :param pos1: 
    :param size1:
    :param pos2:
    :param size2:
    :return: 
    """
    # X overlap
    if pos1[0] <= pos2[0]+size2[0]-1 and pos2[0] <= pos1[0]+size1[0]-1:
        # Y overlap
        if pos1[1] <= pos2[1]+size2[1]-1 and pos2[1] <= pos1[1]+size1[1]-1:
            return True
    return False


def has_values(l):
    """
    Returns True if a nested list ( [[...], [...], ...] ) contains at least
    one truthy value.
    :param l:
    :return:
    """
    for row in l:
        for value in row:
            if value:
                return True
    return False


def blit(l1, l2, x, y):
    """
    Blits l2 to l1 at a given pos, overwriting the original values
    Returns the blitted version of l1
    :param l1:
    :param l2:
    :return:
    """
    if x + len(l2[0]) > len(l1[0]) or y + len(l2) > len(l1):
        raise ValueError('Cannot blit the list where it won\'t fit')
    r = deepcopy(l1)
    for y_offset in range(len(l2)):
        for x_offset in range(len(l2[0])):
            r[y+y_offset][x+x_offset] = l2[y_offset][x_offset]
    return r


#  Exceptions were moved here to avoid circular imports
class BearException(Exception):
    pass


class BearLoopException(BearException):
    pass


class BearLayoutException(BearException):
    pass


class BearECSException(BearException):
    pass


class BearSoundException(BearException):
    pass

class BearJSONException(BearException):
    pass



"""
Entity-component system.

Entities are just the sets of components, nothing more. The major way for them
to be used should be components calling something like
`self.owner.that_other_component.do_stuff` or emitting `ecs_*` events.

The creation of a new Entity is announced by the following event:
`BearEvent(event_type='ecs_create', event_value=entity)`

It is the only event type that uses the actual entity object, not its ID, as the
event_value. When this event is emitted, the entity should be ready to work,
in particular, all its components should be subscribed to the appropriate events
"""
import inspect

from bear_hug.bear_utilities import BearECSException, BearJSONException, \
    rectangles_collide
from bear_hug.widgets import Widget, Listener, deserialize_widget
from bear_hug.event import BearEvent, BearEventDispatcher

from json import dumps, loads


class Entity:
    """
    A root entity class.
    """
    
    def __init__(self, id='Default ID', components=[]):
        self.components = []
        self.id = id
        for component in components:
            self.add_component(component)
        
    def add_component(self, component):
        """
        Adds component to the Entity __dict__
        Raises exception if the Component.name corresponds to one of the already
        existing elements of __dict__ and isn't in self.components. The latter
        check is to allow overwriting components while preventing them from
        overwriting the builtin properties.
        :param component:
        :return:
        """
        if not isinstance(component, Component):
            raise BearECSException('Only Component instance can be added' +
                                   ' as an entity\'s component')
        if component.name not in self.components and \
                component.name in self.__dict__:
            raise BearECSException('Cannot add component' +
                   '{} that shadows builtin attribute'.format(component.name))
        self.__dict__[component.name] = component
        self.components.append(component.name)
        component.owner = self
    
    def remove_component(self, component_name):
        if component_name in self.components:
            del(self.__dict__[component_name])
            self.components.remove(component_name)
        else:
            raise BearECSException('Cannot remove component ' +
                          '{} that Entity doesn\'t have'.format(component_name))
        
    def __repr__(self):
        d = {'id': self.id,
             'components': {x: repr(self.__dict__[x]) for x in self.components}}
        return dumps(d)
        
    
class Component(Listener):
    """
    A root component class.
    
    Component name is expected to be the same between all components of the same
    class. Component inherits from Listener and is therefore able to receive and
    return BearEvents. Of course, it needs the correct subscriptions to actually
    get them.

    `repr(component)` is used for serialization and should generate a valid
    JSON-encoded dict. It should always include a 'class' key which
    should equal the class name for that component and will be used by a
    deserializer to determine what to create. All other keys will be
    deserialized and treated as kwargs to a newly-created object. To define the
    deserialization protocol, JSON dict may also contain keys formatted as
    '{kwarg_name}_type' which should be a string and will be eval-ed as during
    deserialization. Only Python's builtin converters (eg `str`, `int` or
    `float`) are perfectly safe; custom ones are currently unsupported.
    For example, the following is a valid JSON:

    {"class": "TestComponent",
    "x": 5,
    "y": 5,
    "direction": "r",
    "former_owners": ["asd", "zxc", "qwe"],
    "former_owners_type": "set"}

    Its deserialization is equivalent to the following call:
    `x = TestComponent(x=5, y=5, direction='r',
                       former_owners=set(['asd', 'zxc', 'qwe']))`

    The following keys are forbidden: 'name', 'owner', 'dispatcher'. Kwarg
    validity is not controlled except by `Component.__init__()`.

    """
    def __init__(self, dispatcher, name='Root component', owner=None):
        super().__init__()
        if not name:
            raise BearECSException('Cannot create a component without a name')
        if dispatcher and not isinstance(dispatcher, BearEventDispatcher):
            raise BearECSException(f'Attempted to use {type(dispatcher)} as dispatcher')
        self.dispatcher = dispatcher
        self.name = name
        self.owner = None
        self.set_owner(owner)
            
    def set_owner(self, owner):
        """
        Registers a component owner.
        
        This is only useful if the component is passed from one owner to
        another, or if the component is created with the `owner` argument.
        This method calls owner's `add_component`
        :param owner:
        :return:
        """
        if owner:
            if not isinstance(owner, Entity):
                raise BearECSException('Only an Entity can be Component owner')
            owner.add_component(self)
        
    def on_event(self, event):
        """
        Component's event callback.
        :param event:
        :return:
        """
        pass

    def __repr__(self):
        # The most basic serialization. Just states that an Entity has a
        # component of this class, but stores no data
        d = {'class': self.__class__.__name__}
        return dumps(d)

    def __str__(self):
        owner = self.owner.id if self.owner else 'nobody'
        return f'{type(self).__name__} at {id(self)} attached to {owner}'


# Copypasting SO is the only correct way to program
# https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args,
                                                                 **kwargs)
        return cls._instances[cls]


class EntityTracker(Listener, metaclass=Singleton):
    """
    Listens to the ecs_add and ecs_destroy events and keeps track of all the
    currently existing entities.

    This tracker is used for entity lookup, eg by Components that need to find
    all possible entities that fulfill certain criteria
    """

    def __init__(self):
        super().__init__()
        self.entities = {}

    def on_event(self, event):
        if event.event_type == 'ecs_create':
            self.entities[event.event_value.id] = event.event_value
        elif event.event_type == 'ecs_destroy':
            del self.entities[event.event_value]

    def filter_entities(self, key=lambda x: x):
        """
        Return all entities for which key evaluates to True.

        Note that this method returns entity objects themselves, not the IDs.
        :param key: A single-arg callable
        :return:
        """
        if not hasattr(key, '__call__'):
            raise ValueError('EntityTracker requires callable for a key')
        for entity_id in self.entities:
            if key(self.entities[entity_id]):
                yield self.entities[entity_id]


class WidgetComponent(Component):
    """
    Widget as a component.
    
    This component is an ECS wrapper around the Widget object. Since Widgets
    can accept events and it is sometimes reasonable to keep some event logic in
    the Widget instead of Components (ie to keep animation running), its
    `on_event` method simply passes the events to the Widget. It also supports
    `height`, `width` and `size` properties, also by calling widget's ones.
    """
    
    def __init__(self, dispatcher, widget, owner=None):
        if not isinstance(widget, Widget):
            raise TypeError('A widget is not actually a Widget')
        super().__init__(dispatcher=dispatcher, name='widget', owner=owner)
        self.widget = widget
        if self.dispatcher:
            self.dispatcher.register_listener(self, 'tick')

    def on_event(self, event):
        return self.widget.on_event(event)

    @property
    def height(self):
        return self.widget.height

    @property
    def width(self):
        return self.widget.width

    @property
    def size(self):
        return self.widget.size
    
    def __repr__(self):
        d = {'widget': loads(repr(self.widget)),
             'class': self.__class__.__name__}
        return dumps(d)


class PositionComponent(Component):
    """
    A component responsible for positioning Widget on ECSLayout.
    
    It has x and y coordinates, as well as vx and vy speed components.
    Coordinates are given in tiles and speed is in tiles per second.
    """
    def __init__(self, dispatcher, x=0, y=0, vx=0, vy=0, owner=None):
        super().__init__(dispatcher, name='position', owner=owner)
        self._x = x
        self._y = y
        self.x_delay = None
        self.y_delay = None
        self.x_waited = 0
        self.y_waited = 0
        self.vx = vx
        self.vy = vy
        if self.dispatcher:
            dispatcher.register_listener(self, 'tick')

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def vx(self):
        return self._vx

    @vx.setter
    def vx(self, value):
        self._vx = value
        if value != 0:
            self.x_delay = abs(1/self._vx)
        else:
            self.x_delay = None

    @property
    def vy(self):
        return self._vy

    @vy.setter
    def vy(self, value):
        self._vy = value
        if value != 0:
            self.y_delay = abs(1/self._vy)
        else:
            self.y_delay = None

    def move(self, x, y, emit_event=True):
        """
        Move the Entity to a specified position.
        :param x: x
        :param y: y
        :param emit_event: Whether to emit an 'esc_move' event. There are a few
        cases (ie setting the coordinates after the component is created, but
        before the entity is added to the terminal) where this is undesirable.
        :return:
        """
        self._x = x
        self._y = y
        if emit_event:
            self.dispatcher.add_event(BearEvent(event_type='ecs_move',
                                                event_value=(self.owner.id,
                                                             self._x,
                                                             self._y)))

    def relative_move(self, dx, dy, emit_event=True):
        """
        Move the Entity to a specified position.
        :param x: x
        :param y: y
        :param emit_event: Whether to emit an 'esc_move' event. There are a few
        cases (ie setting the coordinates after the component is created, but
        before the entity is added to the terminal) where this is undesirable.
        :return:
        """
        self.move(self.x+dx, self.y+dy, emit_event=emit_event)
        
    def on_event(self, event):
        if event.event_type == 'tick':
            # Move
            if self.vx or self.vy:
                self.x_waited += event.event_value
                self.y_waited += event.event_value
                if self.x_delay and self.x_waited > self.x_delay:
                    new_x = self.x + round(self.x_waited/self.x_delay)\
                        if self.vx > 0\
                        else self.x - round(self.x_waited/self.x_delay)
                    self.x_waited = 0
                else:
                    new_x = self.x
                if self.y_delay and self.y_waited > self.y_delay:
                    new_y = self.y + round(self.y_waited/self.y_delay)\
                        if self.vy > 0\
                        else self.y - round(self.y_waited/self.y_delay)
                    self.y_waited = 0
                else:
                    new_y = self.y
                if not self.x == new_x or not self.y == new_y:
                    self.move(new_x, new_y)

    def __repr__(self):
        d = {'class': self.__class__.__name__,
             'x': self.x,
             'y': self.y,
             'vx': self.vx,
             'vy': self.vy}
        return dumps(d)
                

class DestructorComponent(Component):
    """
    A component responsible for cleanly destroying its entity and everything
    that has to do with it.
    """
    def __init__(self, *args, is_destroying=False, **kwargs):
        super().__init__(*args, name='destructor', **kwargs)
        self.is_destroying = is_destroying
        self.dispatcher.register_listener(self, ['service', 'tick'])
    
    def destroy(self):
        """
        Destruct this component's owner.
        Unsubscribes owner and all its components from the queue and sends
        'ecs_remove'. Then all components are deleted. Entity itself is left at
        the mercy of garbage collector.
        :return:
        """
        self.dispatcher.add_event(BearEvent('ecs_destroy', self.owner.id))
        self.is_destroying = True
        # Destroys item on the 'tick_over', so that all
        # existing events involving owner (including 'ecs_remove)' are processed
        # normally, but unsubscribes it right now to prevent new ones from forming
        for component in self.owner.components:
            if component != self.name:
                self.dispatcher.unregister_listener(
                    self.owner.__dict__[component])
    
    def on_event(self, event):
        if self.is_destroying and event.event_type == 'service' and event.event_value == 'tick_over':
            # owner.components stores IDs, not component objects themselves.
            # Those are available only from owner.__dict__
            victims = [x for x in self.owner.components]
            for component in victims:
                if component is not self.name:
                    self.dispatcher.unregister_listener(self.owner.__dict__[component])
                    self.owner.remove_component(component)
            self.dispatcher.unregister_listener(self)
            self.owner.remove_component(self.name)

    def __repr__(self):
        d = {'class': self.__class__.__name__,
             'is_destroying': self.is_destroying} # Could be saved right in the middle of destruction
        return dumps(d)


class CollisionComponent(Component):
    """
    A component responsible for processing collisions of this object
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, name='collision', **kwargs)
        self.dispatcher.register_listener(self, 'ecs_collision')

    def on_event(self, event):
        if event.event_type == 'ecs_collision':
            if event.event_value[0] == self.owner.id:
                self.collided_into(event.event_value[1])
            elif event.event_value[1] == self.owner.id:
                self.collided_by(event.event_value[0])

    def collided_into(self, entity):
        pass

    def collided_by(self, entity):
        pass


class WalkerCollisionComponent(CollisionComponent):
    """
    A collision component that, upon colliding into something impassable,
    moves the entity to where it came from. Expects both entities involved to
    have a PassabilityComponent
    """

    def collided_into(self, entity):
        if entity is not None:
            other = EntityTracker().entities[entity]
            if 'passability' in self.owner.__dict__ and 'passability' in other.__dict__:
                if rectangles_collide((self.owner.position.x +
                                       self.owner.passability.shadow_pos[0],
                                       self.owner.position.y +
                                       self.owner.passability.shadow_pos[1]),
                                      self.owner.passability.shadow_size,
                                      (other.position.x +
                                       other.passability.shadow_pos[0],
                                       other.position.y +
                                       other.passability.shadow_pos[1]),
                                      other.passability.shadow_size):
                    self.owner.position.relative_move(
                        self.owner.position.last_move[0] * -1,
                        self.owner.position.last_move[1] * -1)
        else:
            # Processing collisions with screen edges without involving passability
            self.owner.position.relative_move(
                self.owner.position.last_move[0] * -1,
                self.owner.position.last_move[1] * -1)


class PassingComponent(Component):
    """
    A component responsible for knowing whether items can or cannot be walked
    through.

    Unlike collisions of eg projectiles, walkers can easily collide with screen
    items and each other provided they are "behind" or "ahead" of each other. To
    check for that, PassingComponent stores a sort of hitbox (basically the
    projection on the surface, something like lowest three rows for a
    human-sized object). Then, WalkerCollisionComponent uses those to define
    if walk attempt was unsuccessful.

    All entities that do not have this component are assumed to be passable.
    """

    def __init__(self, *args, shadow_pos=(0, 0), shadow_size=None, **kwargs):
        super().__init__(*args, name='passability', **kwargs)
        self.shadow_pos = shadow_pos
        self._shadow_size = shadow_size

    @property
    def shadow_size(self):
        # TODO: remove the ugly shadow size hack
        # The idea is that shadow size can be set to owner's widget size by
        # default. The only issue is that owner may not be set, or may not have
        # a widget yet, when this component is created. Thus, this hack.
        # Hopefully no one will try and walk into the object before it is shown
        # on screen. Alas, it requires calling a method for a frequently used
        # property and is generally pretty ugly. Remove this if I ever get to
        # optimizing and manage to think of something better.
        if self._shadow_size is None:
            self._shadow_size = self.owner.widget.size
        return self._shadow_size

    def __repr__(self):
        d = {'class': self.__class__.__name__,
             'shadow_size': self._shadow_size,
             'shadow_pos': self.shadow_pos}
        return dumps(d)


class DecayComponent(Component):
    """
    Attaches to an entity and destroys it when conditions are met.

    Currently supported destroy conditions are 'keypress' and 'timeout'. If the
    latter is set, you can supply the lifetime (defaults to 1.0 sec)
    """

    def __init__(self, *args, destroy_condition='keypress', lifetime=1.0, age=0,
                 **kwargs):
        super().__init__(*args, name='decay', **kwargs)
        if destroy_condition == 'keypress':
            self.dispatcher.register_listener(self, 'key_down')
        elif destroy_condition == 'timeout':
            self.dispatcher.register_listener(self, 'tick')
            self.lifetime = lifetime
            self.age = age
        else:
            raise ValueError(
                f'destroy_condition should be either keypress or timeout')
        self.destroy_condition = destroy_condition

    def on_event(self, event):
        if self.destroy_condition == 'keypress' and event.event_type == 'key_down':
            self.owner.destructor.destroy()
        elif self.destroy_condition == 'timeout' and event.event_type == 'tick':
            self.age += event.event_value
            if self.age >= self.lifetime:
                self.owner.destructor.destroy()

    def __repr__(self):
        return dumps({'class': self.__class__.__name__,
                      'destroy_condition': self.destroy_condition,
                      'lifetime': self.lifetime,
                      'age': self.age})


def deserialize_component(serial, dispatcher):
    """
    Provided a JSON string, creates a necessary object.
    
    Does not subscribe a component to anything (which can be done either by a
    caller or in the ComponentSubclass.__init__) or assign it to any Entity
    (which is probably done within `deserialize_entity`)
    :param json_string:
    :param dispatcher:
    :return:
    """
    if isinstance(serial, str):
        d = loads(serial)
    elif isinstance(serial, dict):
        d = serial
    else:
        raise BearJSONException(f'Attempting to deserialize {type(serial)} to Component')
    for forbidden_key in ('name', 'owner', 'dispatcher'):
        if forbidden_key in d.keys():
            raise BearJSONException(f'Forbidden key {forbidden_key} in component JSON')
    if 'class' not in d:
        raise BearJSONException('No class provided in component JSON')
    # Only builtins supported for converters. Although custom converters could
    # be provided like with classes, IMO this way is safer
    converters = {}
    for key in d:
        if '_type' in key:
            converters[key[:-5]] = globals()['__builtins__'][d[key]]
    types = [x for x in d if '_type' in x]
    for t in types:
        del(d[t])
    # Try to get the Component class from where the function was imported, or
    # the importers of *that* frame. Without this, the function would only see
    # classes from this very file, or ones imported into it, and that would
    # break the deserialization of custom components.
    class_var = None
    for frame in inspect.getouterframes(inspect.currentframe()):
        if d['class'] in frame.frame.f_globals:
            class_var = frame.frame.f_globals[d['class']]
            break
    del frame
    if not class_var:
        raise BearJSONException(f"Class name {class_var} not imported anywhere in the frame stack")
    if not issubclass(class_var, Component):
        raise BearJSONException(f"Class name {class_var}mapped to something other than a Component subclass")
    kwargs = {}
    for key in d:
        if key == 'class':
            continue
        if key in converters:
            kwargs[key] = converters[key](d[key])
        elif key == 'widget':
            w = deserialize_widget(d['widget'])
            #TODO: subscribe widgets to events other than 'tick' in deserialization
            dispatcher.register_listener(w, 'tick')
            kwargs['widget'] = w
        else:
            kwargs[key] = d[key]
    return class_var(dispatcher, **kwargs)


def deserialize_entity(serial, dispatcher):
    """Load the entity from JSON string or dict.
    
    Does not subscribe a new entity to anything or emit `bear_create` events;
    this should be done by a caller."""
    if isinstance(serial, str):
        d = loads(serial)
    elif isinstance(serial, dict):
        d = serial
    else:
        raise BearJSONException(f'Attempting to deserialize {type(serial)} to Entity')
    components = [deserialize_component(d['components'][x], dispatcher)
                  for x in d['components']]
    return Entity(id=d['id'], components=components)
    
    




"""
Various useful Widget and Listener classes
These widgets and listeners are usable outside the ECS and should be sufficient
for simpler games and apps. However, for the sake of clearer architecture,
entities are recommended.
"""

# TODO: rebuild JSON serialization system to avoid dumping Widgets altogether
# This is essentially mixing up model and view, which is generally bad. However,
# it has marginal benefits (for dumping procedurally generated and/or edited
# widgets within entities. Although actually it exists mostly because I want a
# functional dumping system right now, and disentangling them would make
# WidgetComponent serializer a huge mess.

import inspect

from bear_hug.bear_hug import BearTerminal
from bear_hug.bear_utilities import shapes_equal, blit, copy_shape, slice_nested,\
    BearException, BearLayoutException, BearJSONException
from bear_hug.event import BearEvent

from collections import deque
from json import dumps, loads
from time import time


def deserialize_widget(serial, atlas=None):
    """
    Provided a JSON string, return a widget it encodes.
    :param json_string:
    :param dispatcher:
    :return:
    """
    # TODO: support getting chars and colors for deserialization from atlas
    # In animation, it is achieved by storing source IDs in __init__. Not
    # certain if it'd be safe for Widgets with all their complexity
    
    if isinstance(serial, str):
        d = loads(serial)
    elif isinstance(serial, dict):
        d = serial
    else:
        raise BearJSONException(f'Attempting to deserialize {type(serial)} to Widget')
    for forbidden_key in ('name', 'owner', 'dispatcher'):
        if forbidden_key in d.keys():
            raise BearJSONException(f'Forbidden key {forbidden_key} in widget JSON')
    if 'class' not in d:
        raise BearJSONException('No class provided in component JSON')
    # Only builtins supported for converters. Although custom converters could
    # be provided like with classes, IMO this way is safer
    converters = {}
    for key in d:
        if '_type' in key:
            converters[key[:-5]] = globals()['__builtins__'][d[key]]
    types = [x for x in d if '_type' in x]
    for t in types:
        del(d[t])
    # Try to get the Widget subclass from where the function was imported, or
    # the importers of *that* frame. Without this, the function would only see
    # classes from this very file, or ones imported into it, and that would
    # break the deserialization of custom components.
    class_var = None
    for frame in inspect.getouterframes(inspect.currentframe()):
        if d['class'] in frame.frame.f_globals:
            class_var = frame.frame.f_globals[d['class']]
            break
    del frame
    if not class_var:
        raise BearJSONException(f"Class name {d['class']} not imported anywhere in frame stack")
    if not issubclass(class_var, Widget):
        raise BearJSONException(f"Class name {d['class']}mapped to something other than a Widget subclass")
    kwargs = {}
    for key in d:
        if key in {'class', 'chars', 'colors'}:
            continue
        elif key in converters:
            kwargs[key] = converters[key](d[key])
        elif key == 'animation':
            # animation deserializer will raise exception if atlas is not supplied
            kwargs['animation'] = deserialize_animation(d['animation'], atlas)
        else:
            kwargs[key] = d[key]
    if 'chars' in d:
        # Chars and colors are not kwargs
        return class_var(chars=[[char for char in x] for x in d['chars']],
                         colors=[x.split(',') for x in d['colors']],
                         **kwargs)
    else:
        # Some classes, eg animation widgets, do not dump chars and colors
        return class_var(**kwargs)


def deserialize_animation(serial, atlas=None):
    """
    Deserialize an animation from JSON dump
    :param serial:
    :return:
    """
    d = loads(serial)
    if d['storage_type'] == 'atlas':
        if not atlas:
            raise BearJSONException('Animation storage type set to atlas, but atlas was not supplied')
        return Animation(frames=[atlas.get_element(x) for x in d['frame_ids']],
                         fps=d['fps'])
    elif d['storage_type'] == 'dump':
        return Animation(frames=[[[[char for char in x] for x in frame[0]],
                                  [x.split(',') for x in frame[1]]]
                                 for frame in d['frames']],
                         fps=d['fps'])
    else:
        raise BearJSONException(f"Incorrect Animation storage_type: {d['storage_type']}")


class Widget:
    """
    The base class for things that can be placed on the terminal.
    This class is inactive and is intended to be either inherited from or used
    for purely decorative non-animated objects. Event processing and animations
    are covered by its subclasses; while it has `on_event()` method, it does
    nothing. This allows Widgets to work without subscribing to the queue and
    saves some tacts on not redrawing them unless the Widget itself considers it
    necessary.

    Accepted parameters:
    `chars`: a list of unicode characters
    `colors`: a list of colors. Anything that is accepted by terminal.color()
    goes here (a color name or a 0xAARRGGBB/0xRRGGBB/0xRGB/0xARGB integer).

    `chars` and `colors` should be exactly the same shape, otherwise the
    BearException is raised.
    
    Widgets can be serialized into JSON similarly to Components and Entities.
    `repr(widget)` is used for serialization and should generate a valid
    JSON-encoded dict. It should always include a 'class' key which
    should equal the class name for that component and will be used by a
    deserializer to determine what to create. `chars` and `colors` keys are also
    necessary. They should encode widget's chars and colors as arrays of strings
    and each of these strings should be a list of values for
    chars' and colors' inner lists (str-converted chars and str-converted
    `#ffffff`-type colors; comma-separated for colors).
    
    All other keys will be deserialized and treated as kwargs to a newly-created
    object. To define the deserialization protocol, JSON dict may also contain
    keys formatted as '{kwarg_name}_type' which should be a string and will be
    eval-ed as during deserialization. Only Python's builtin converters (eg
    `str`, `int` or `float`) are safe; custom ones are currently unsupported.
    For example, the following is a valid JSON:

    ```
    {"class": "MyWidget",
    "chars": ["b,b,b", "a,b,a", "b,a,b"],
    "colors": ["#fff,#fff,#fff", "#000,#fff,#000", "#fff,#000,#fff"],
    "former_owners": ["asd", "zxc", "qwe"],
    "former_owners_type": "set"}
    ```

    Its deserialization is equivalent to the following call:
    ```
    `x = MyWidget(chars=[['bbb'],
                         ['aba'],
                         ['bab']],
                  colors=[['#fff','#fff','#fff'],
                          ['#000','#fff','#000'],
                          ['#fff','#000','#fff']],
                  former_owners=set(['asd, 'zxc', 'qwe']))
    ```

    The following keys are forbidden: `parent` and `terminal`. Kwarg validity
    is not controlled except by `Component.__init__()`.
    """
    #TODO: maybe support background colour after all?
    
    def __init__(self, chars, colors):
        if not isinstance(chars, list) or not isinstance(colors, list):
            raise BearException('Chars and colors should be lists')
        if not shapes_equal(chars, colors):
            raise BearException('Chars and colors should have the same shape')
        self.chars = chars
        self.colors = colors
        # A widget may want to know about the terminal it's attached to
        self._terminal = None
        # Or a parent
        self._parent = None
        
    def on_event(self, event):
        # Root widget does not raise anything here, because Widget() can be
        # erroneously subscribed to a queue. While useless, that's not really a
        # fatal error.
        pass
    
    @property
    def terminal(self):
        return self._terminal
    
    @terminal.setter
    def terminal(self, value):
        if value and not isinstance(value, BearTerminal):
            raise BearException('Only a BearTerminal can be set as ' +
                                'Widget.terminal')
        self._terminal = value
        
    @property
    def parent(self):
        return self._parent
    
    @parent.setter
    def parent(self, value):
        if value and not isinstance(value, (Widget, BearTerminal)):
            raise BearException(
                'Only a widget or terminal can be a widget\'s parent')
        self._parent = value
        
    @property
    def height(self):
        return len(self.chars)
    
    @property
    def width(self):
        return len(self.chars[0])
    
    @property
    def size(self):
        return len(self.chars[0]), len(self.chars)
        
    def flip(self, axis):
        """
        Flip a widget along one of the axes.
        
        Rotates chars and colors either horizontally ('x' or 'horizontal') or
        vertically ('y' or 'vertical'). Note that this method has **extremely**
        limited uses: first, it only affects chars and colors *as they are now*.
        If later the widget gets updated via animation, updating label text,
        Layout's children being redrawn, etc., it will be un-flipped again.
        Second, most ASCII-art does not take it well. You're probably safe with
        something almost symmetrical, like background garbage and such, but for
        complex images it's better to provide both left and right versions.
        
        Unlike raster and vector graphics, there is no general way to flip an
        ASCII image programmatically. Except, of course, flipping tiles
        themselves which I find aesthetically unacceptable for my projects.
        
        :param axis:
        :return:
        """
        if axis in ('x', 'horizontal'):
            self.chars = [self.chars[x][::-1] for x in range(len(self.chars))]
            self.colors = [self.colors[x][::-1] for x in range(len(self.colors))]
        elif axis in ('y', 'vertical'):
            self.chars = self.chars[::-1]
            self.colors = self.colors[::-1]
            
    def __repr__(self):
        char_strings = [''.join(x) for x in self.chars]
        for string in char_strings:
            string.replace('\"', '\u0022"').replace('\\', '\u005c')
        d = {'class': self.__class__.__name__,
             'chars': char_strings,
             'colors': [','.join(x) for x in self.colors]}
        return dumps(d)


class SwitchingWidget(Widget):
    """
    A widget that can contain a collection of chars/colors pairs and switch
    them on command. These char/color pairs should all be the same shape.
    Does not do any transition animations.
    """
    
    def __init__(self, chars=None, colors=None, images_dict=None, initial_image=None):
        # Chars and colors are not used anywhere; they are included simply for
        # the compatibility with serialization. Actual chars and colors of the
        # SwitchingWidget are set to `images_dict[initial_image]` upon creation
        test_shape = None
        for image in images_dict:
            # Checking if the image is from JSON (each line is a string) or a
            # correct list-of-lists. If it's former, converts
            if isinstance(images_dict[image][0][0], str):
                images_dict[image][0] = [[char for char in x]
                                         for x in images_dict[image][0]]
                images_dict[image][1] = [list(x.split(','))
                                         for x in images_dict[image][1]]
            if not shapes_equal(images_dict[image][0], images_dict[image][1]):
                raise BearException(
                    f'Chars and colors of different shape for image ID {image} in SwitchingWidget')
            if not test_shape:
                test_shape = (len(images_dict[image][0]),
                              len(images_dict[image][0][0]))
            elif len(images_dict[image][0]) != test_shape[0] or \
                    len(images_dict[image][0][0]) != test_shape[1]:
                raise BearException(
                    f'Image {image} in SwitchingWidget has incorrect size')
        if not initial_image:
            raise BearException('Initial image not set for SwitchingWidget')
        super().__init__(*images_dict[initial_image])
        self.images = images_dict
        self.current_image = initial_image
    
    def switch_to_image(self, image_id):
        if image_id != self.current_image:
            try:
                self.chars = self.images[image_id][0]
                self.colors = self.images[image_id][1]
                self.current_image = image_id
            except KeyError:
                raise BearException(
                    f'Attempting to switch to incorrect image ID {image_id}')
            
    def __repr__(self):
        d = {'class': self.__class__.__name__,
             'initial_image': self.current_image}
        images = {}
        for image in self.images:
            images[image] = []
            # Seems to work without any complex workarounds for screening
            images[image].append([''.join(x)#.replace('"', '\u0022"').
                                            # replace('\\', '\u005c')
                                  for x in self.images[image][0]])
            images[image].append([','.join(x) for x in self.images[image][1]])
        d['images_dict'] = images
        return dumps(d)
        
        
class Layout(Widget):
    """
    A widget that can add others as its children.
    All children get drawn to its `chars` and are thus displayed on a single
    bearlibterminal layer. The layout does not explicitly pass events to its
    children, they are expected to subscribe to event queue by themselves.
    Children are allowed to overlap, but in that case the most recent one's char
    is actually drawn.
    The Layout is initialized with a single child, which is given chars and
    colors provided at Layout creation. This child is available as l.children[0]
    or as l.background
    The Layout automatically redraws itself on `tick` event, whether its
    children have updated or not.
    
    Does not currently support JSON serialization
    """
    def __init__(self, chars, colors):
        super().__init__(chars, colors)
        self.children = []
        # For every position, remember all the widgets that may want to place
        # characters in it, but draw only the latest one
        self._child_pointers = copy_shape(self.chars, None)
        # copy_shape does not work with lists correctly, so.
        for line in range(len(self._child_pointers)):
            for char in range(len(self._child_pointers[0])):
                self._child_pointers[line][char] = []
        self.child_locations = {}
        # The widget with Layout's chars and colors is created and added to the
        # Layout as the first child. It is done even if both are empty, just in
        # case someone wants to add background later
        w = Widget(self.chars, self.colors)
        self.add_child(w, pos=(0, 0))
    
    @property
    def terminal(self):
        return self._terminal
    
    # This setter propagates the terminal value to all the Layout's children.
    # It's necessary because some of them may be added before placing Layout on
    # the screen and thus end up terminal-less.
    @terminal.setter
    def terminal(self, value):
        if value and not isinstance(value, BearTerminal):
            raise BearException('Only BearTerminal can be added as terminal')
        self._terminal = value
        for child in self.children:
            child.terminal = value
        
    # Operations on children
    def add_child(self, child, pos, skip_checks = False):
        """
        Add a widget as a child at a given (relative) position.
        The child has to be a Widget or a Widget subclass that haven't yet been
        added to this Layout and whose dimensions are less than or equal to the
        Layout's
        :param child:
        :return:
        """
        if not isinstance(child, Widget):
            raise BearLayoutException('Cannot add non-Widget to a Layout')
        if child in self.children and not skip_checks:
            raise BearLayoutException('Cannot add the same widget to layout twice')
        if len(child.chars) > len(self._child_pointers) or \
                len(child.chars[0]) > len(self._child_pointers[0]):
            raise BearLayoutException('Cannot add child that is bigger than a Layout')
        if len(child.chars) + pos[1] > len(self._child_pointers) or \
                len(child.chars[0]) + pos[0] > len(self._child_pointers[0]):
            raise BearLayoutException('Child won\'t fit at this position')
        if child is self:
            raise BearLayoutException('Cannot add Layout as its own child')
        if not skip_checks:
            self.children.append(child)
        self.child_locations[child] = pos
        child.terminal = self.terminal
        child.parent = self
        for y in range(len(child.chars)):
            for x in range(len(child.chars[0])):
                self._child_pointers[pos[1] + y][pos[0] + x].append(child)

    def remove_child(self, child, remove_completely=True):
        """
        Remove a child from a Layout.
        :param child: the child to remove
        :param remove_completely: if False, the child is only removed from the
        screen, but remains in the children list. This is not intended to be
        used and is included only to prevent self.move_child from messing with
        child order.
        :return:
        """
        if child not in self.children:
            raise BearLayoutException('Layout can only remove its child')
        # process pointers
        for y in range(len(child.chars)):
            for x in range(len(child.chars[0])):
                self._child_pointers[self.child_locations[child][1] + y] \
                        [self.child_locations[child][0] + x].remove(child)
        if remove_completely:
            del(self.child_locations[child])
            self.children.remove(child)
            child.terminal = None
            child.parent = None
    
    def move_child(self, child, new_pos):
        self.remove_child(child, remove_completely=False)
        self.add_child(child, pos=new_pos, skip_checks=True)
    
    # BG's chars and colors are not meant to be set directly
    @property
    def background(self):
        return self.children[0]
    
    @background.setter
    def background(self, value):
        if not isinstance(value, Widget):
            raise BearLayoutException('Only Widget can be added as background')
        if not shapes_equal(self.chars, value.chars):
            # chars and colors are always the same size
            raise BearLayoutException('Wrong Layout background size')
        for row in range(len(self.chars)):
            for column in range(len(self.chars[0])):
                self._child_pointers[row][column][0] = value
                # self.colors[row][column][0] = value
        del self.child_locations[self.children[0]]
        self.child_locations[value] = (0, 0)
        self.children[0] = value
        
    def _rebuild_self(self):
        """
        Build fresh chars and colors for self
        :return:
        """
        # TODO: Support needs_redraw like in ECSLayout
        chars = copy_shape(self.chars, ' ')
        colors = copy_shape(self.colors, None)
        for line in range(len(chars)):
            for char in range(len(chars[0])):
                for child in self._child_pointers[line][char][::-1]:
                    # Addressing the correct child position
                    c = child.chars[line-self.child_locations[child][1]] \
                        [char - self.child_locations[child][0]]
                    if c != ' ':
                        # Spacebars are used as empty space and are transparent
                        chars[line][char] = c
                        break
                colors[line][char] = \
                    child.colors[line - self.child_locations[child][1]] \
                    [char - self.child_locations[child][0]]
        self.chars = chars
        self.colors = colors
    
    def on_event(self, event):
        """
        The Layout redraws itself on every frame
        :return:
        """
        if event.event_type == 'service' and event.event_value == 'tick_over':
            self._rebuild_self()
            if isinstance(self.parent, BearTerminal):
                self.terminal.update_widget(self)
    
    #Service
    def get_absolute_pos(self, relative_pos):
        """
        Get an absolute position (in terminal coordinates) for any location
        within self.
        :param relative_pos:
        :return:
        """
        self_pos = self.terminal.widget_locations(self).pos
        return self_pos[0]+relative_pos[0], self_pos[1]+relative_pos[1]
    
    def __repr__(self):
        raise BearException('Layout does not support __repr__ serialization')


class ScrollBar(Widget):
    """
    A scrollbar to be used with ScrollableLayout. On creation accepts the
    following parameters:
    `orientation` -- one of 'vertical' or 'horizontal'
    `length` -- int
    'colors' -- two BLT-compatible colors for background and the moving thingy
    
    Does not accept input, does not support serialization
    """
    def __init__(self, orientation='vertical', length=10,
                 colors=('gray', 'white')):
        if orientation not in ('vertical', 'horizontal'):
            raise BearException(
                'Orientation must be either vertical or horizontal')
        if orientation == 'vertical':
            # TODO: custom chars in ScrollBar
            chars = [['#'] for _ in range(length)]
        else:
            chars = [['#' for _ in range(length)]]
        self.length = length
        self.orientation = orientation
        self.bg_color = colors[0]
        self.bar_color = colors[1]
        colors = copy_shape(chars, self.bg_color)
        super().__init__(chars, colors)
        
    def show_pos(self, position, percentage):
        """
        Move the scrollbar.
        :param position: Float. The position of the top (or left) side of the
        scrollbar, as part of its length
        :param percentage: Float. The lengths of the scrollbar, as part of the
        total bar length
        :return:
        """
        # Not really effective, but still quicker than Layout would be
        # Single-widget bar gets redrawn only when called, while a Layout
        # would've redrawn every tick
        start = round(self.length*position)
        width = round(self.length*percentage)
        self.colors = copy_shape(self.chars, self.bg_color)
        if self.orientation == 'vertical':
            for i in range(start, start+width):
                self.colors[i][0] = self.bar_color
        else:
            for i in range(start, start+width):
                self.colors[0][i] = self.bar_color
                
    def __repr__(self):
        raise BearException('ScrollBar does not support __repr__ serialization')


class ScrollableLayout(Layout):
    """
    A Layout that can show only a part of its surface.
    
    Like a Layout, accepts `chars` and `colors` on creation, which should be the
    size of the entire layout, not the visible area. The latter is initialized
    by `view_pos` and `view_size` arguments.
    
    Works by overloading _rebuild_self to only show a part of child_pointers
    
    Does not support JSON serialization
    """
    def __init__(self, chars, colors,
                 view_pos=(0, 0), view_size=(10, 10)):
        super().__init__(chars, colors)
        if not 0 <= view_pos[0] <= self.width - view_size[0] \
                or not 0 <= view_pos[1] <= self.height - view_size[1]:
            raise BearLayoutException('Initial viewpoint outside ' +
                                      'ScrollableLayout')
        if not 0 < view_size[0] <= len(chars[0]) \
                or not 0 < view_size[1] <= len(chars):
            raise BearLayoutException('Invalid view field size')
        self.view_pos = view_pos[:]
        self.view_size = view_size[:]
        self._rebuild_self()
    
    def _rebuild_self(self):
        """
        Same as `Layout()._rebuild_self`, but all child positions are also
        offset by `view_pos`. Obviously, only `view_size[1]` lines
        `view_size[0]` long are set as `chars` and `colors`.
        :return:
        """
        chars = [[' ' for x in range(self.view_size[0])] \
                 for y in range(self.view_size[1])]
        colors = copy_shape(chars, None)
        for line in range(self.view_size[1]):
            for char in range(self.view_size[0]):
                for child in self._child_pointers[self.view_pos[1]+line] \
                                     [self.view_pos[0] + char][::-1]:
                    # Addressing the correct child position
                    c = child.chars[self.view_pos[1] + line-self.child_locations[child][1]] \
                        [self.view_pos[0] + char-self.child_locations[child][0]]
                    if c != ' ':
                        # Spacebars are used as empty space and are transparent
                        chars[line][char] = c
                        break
                colors[line][char] = \
                    child.colors[self.view_pos[1] + line - self.child_locations[child][1]] \
                    [self.view_pos[0] + char - self.child_locations[child][0]]
        self.chars = chars
        self.colors = colors
    
    def resize_view(self, new_size):
        # TODO: support resizing view.
        # This will require updating the pointers in terminal or parent layout
        pass
    
    def scroll_to(self, pos):
        """
        Move field of view to `pos`.
        
        Raises `BearLayoutException` on incorrect position
        :param pos: tuple of ints
        :return:
        """
        if not (len(pos) == 2 and all((isinstance(x, int) for x in pos))):
            raise BearLayoutException('Field of view position should be 2 ints')
        if not 0 <= pos[0] <= len(self._child_pointers[0]) - self.view_size[0] \
                or not 0 <= pos[1] <= len(self._child_pointers)-self.view_size[1]:
            raise BearLayoutException('Scrolling to invalid position')
        self.view_pos = pos
    
    def scroll_by(self, shift):
        """
        Move field of view by `shift[0]` to the right and by `shift[1]` down.
        
        Raises `BearLayoutException` on incorrect position
        :param shift: tuple of ints
        :return:
        """
        pos = (self.view_pos[0] + shift[0], self.view_pos[1] + shift[1])
        self.scroll_to(pos)
    
    def __repr__(self):
        raise BearException('ScrollableLayout does not support __repr__ serialization')
    

class InputScrollable(Layout):
    """
    A ScrollableLayout wrapper that accepts input events and supports the usual
    scrollable view bells and whistles. Like ScrollableLayout, accepts chars and
    colors the size of the *entire* layout and inits visible area using view_pos
    and view_size.
    
    If bottom_bar and/or right_bar is set to True, it will be made one char
    bigger than view_size in the corresponding dimension to add ScrollBar.
    
    Can be scrolled by arrow keys.
    
    Does not support JSON serialization
    """
    def __init__(self, chars, colors, view_pos=(0, 0), view_size=(10, 10),
                 bottom_bar=False, right_bar=False):
        # Scrollable is initalized before self to avoid damaging it by the
        # modified view_size (in case of scrollbars)
        scrollable = ScrollableLayout(chars, colors, view_pos, view_size)
        size = list(view_size)
        ch = slice_nested(chars, view_pos, size)
        co = slice_nested(colors, view_pos, size)
        # Is there something more reasonable to add as ScrollableLayout BG?
        # It shouldn't be shown anyway
        if bottom_bar:
            ch.append(copy_shape(ch[0], ch[0][0]))
            co.append(copy_shape(co[0], co[0][0]))
        if right_bar:
            for x in ch:
                x.append(' ')
            for x in co:
                x.append('white')
        # While True, can add children to self. Otherwise they are passed to
        # self.scrollable
        self.building_self = True
        super().__init__(ch, co)
        self.scrollable = scrollable
        self.add_child(self.scrollable, pos=(0, 0))
        # Need to rebuild now to let bars know the correct height and width
        self._rebuild_self()
        if right_bar:
            self.right_bar = ScrollBar(orientation='vertical',
                                       length=self.height)
            self.add_child(self.right_bar, pos=(self.width-1, 0))
        else:
            self.right_bar = None
        if bottom_bar:
            self.bottom_bar = ScrollBar(orientation='horizontal',
                                        length=self.width)
            self.add_child(self.bottom_bar, pos=(0, self.height-1))
        self.building_self = False
        
    def on_event(self, event):
        if event.event_type == 'key_down':
            scrolled = False
            if event.event_value == 'TK_DOWN' and \
              self.scrollable.view_pos[1] + self.scrollable.view_size[1]\
                    < len(self.scrollable._child_pointers):
                self.scrollable.scroll_by((0, 1))
                scrolled = True
            elif event.event_value == 'TK_UP' and \
             self.scrollable.view_pos[1] > 0:
                self.scrollable.scroll_by((0, -1))
                scrolled = True
            elif event.event_value == 'TK_RIGHT' and \
              self.scrollable.view_pos[0] + self.scrollable.view_size[0]\
                    < len(self.scrollable._child_pointers[0]):
                self.scrollable.scroll_by((1, 0))
                scrolled = True
            elif event.event_value == 'TK_LEFT' and \
              self.scrollable.view_pos[0] > 0:
                self.scrollable.scroll_by((-1, 0))
                scrolled = True
            elif event.event_type == 'TK_SPACE':
                self.scrollable.scroll_to((0, 0))
                scrolled = True
            if scrolled:
                if self.right_bar:
                    self.right_bar.show_pos(
                        self.scrollable.view_pos[1] /
                            len(self.scrollable._child_pointers),
                        self.scrollable.view_size[0] /
                            len(self.scrollable._child_pointers))
        super().on_event(event)

    def add_child(self, child, pos, skip_checks=False):
        if not self.building_self:
            self.scrollable.add_child(child, pos, skip_checks)
        else:
            super().add_child(child, pos, skip_checks)
            
    def __repr__(self):
        raise BearException('InputScrollable does not support __repr__ serialization')
            

# Animations and other complex decorative Widgets
class Animation:
    """
    A data class for animation, ie the sequence of the frames
    
    Animation is serialized to JSON, preserving fps and either frame dumps
    (similarly to widget chars and colors) or frame image IDs. For the latter to
    work, these IDs should be provided during Animation creation via an optional
    `frame_ids` kwarg. The deserializer will then use them with whichever atlas
    is supplied to create the animation.
    
    Since this class has no idea of atlases and is unaware whether it was
    created with the same atlas as deserializer will use (which REALLY should be
    the same, doing otherwise is just asking for trouble), frame ID validity is
    not checked until deserialization and, if incorrect, are not guaranteed to
    work.
    """
    def __init__(self, frames, fps, frame_ids=None):
        if not all((shapes_equal(x[0], frames[0][0]) for x in frames[1:])) \
                or not all(
                (shapes_equal(x[1], frames[0][1]) for x in frames[1:])):
            raise BearException('Frames should be equal size')
        if frame_ids:
            if len(frame_ids) != len(frames):
                raise BearJSONException('Incorrect frame_ids length during Animation creation')
            else:
                self.frame_ids = frame_ids
        self.frames = frames
        self.fps = fps # For deserialization
        self.frame_time = 1 / fps

    def __len__(self):
        return len(self.frames)
    
    def __repr__(self):
        d = {'fps': self.fps}
        if hasattr(self, 'frame_ids'):
            d['storage_type'] = 'atlas'
            d['frame_ids'] = 'frame_ids'
        else:
            frames_dump = []
            for frame in self.frames:
                char_strings = [''.join(x) for x in frame[0]]
                for string in char_strings:
                    string.replace('\"', '\u0022"').replace('\\', '\u005c')
                colors_dump = [','.join(x) for x in frame[1]]
                frames_dump.append([char_strings, colors_dump])
            d['frames'] = frames_dump
            d['storage_type'] = 'dump'
        return dumps(d)
        

class SimpleAnimationWidget(Widget):
    """
    A simple animated widget that cycles through the frames.

    Accepts two parameters on creation:
    `frames` an iterable of (chars, colors) tuples. These should all be
    the same size
    `fps` frames per second.
    `emit_ecs`: whether to emit ecs_update events on every frame. Useless for
    widgets outside ecs system, but those on ECSLayout are not redrawn unless
    this event is emitted or something else causes ECSLayout to redraw
    """
    
    def __init__(self, animation, emit_ecs=True):
        if not isinstance(animation, Animation):
            raise BearException(
                'Only Animation instance can be used in SimpleAnimationWidget')
        self.animation = animation
        super().__init__(*animation.frames[0])
        self.running_index = 0
        self.have_waited = 0
        self.emit_ecs = emit_ecs
    
    def on_event(self, event):
        if event.event_type == 'tick':
            self.have_waited += event.event_value
            if self.have_waited >= self.animation.frame_time:
                self.running_index += 1
                if self.running_index >= len(self.animation):
                    self.running_index = 0
                self.chars = self.animation.frames[self.running_index][0]
                self.colors = self.animation.frames[self.running_index][1]
                self.have_waited = 0
                if self.emit_ecs:
                    return BearEvent(event_type='ecs_update')
        elif self.parent is self.terminal and event.event_type == 'service' \
                and event.event_value == 'tick_over':
            # This widget is connected to the terminal directly and must update
            # itself without a layout
            self.terminal.update_widget(self)

    def __repr__(self):
        d = {'class': self.__class__.__name__,
             'animation': repr(self.animation),
             'emit_ecs': self.emit_ecs}
        return dumps(d)


class MultipleAnimationWidget(Widget):
    """
    A widget that is able to display multiple animations.
    """
    def __init__(self, animations, initial_animation,
                 emit_ecs=True, cycle=False):
        # Check the animations' validity
        if not isinstance(animations, dict) or \
                any((not isinstance(x, Animation) for x in animations.values())):
            raise BearException(
                'Only dict of Animations acceptable for MultipleAnimationWidget')
        if any((not isinstance(x, str) for x in animations)):
            raise BearException('Animation names should be strings')
        if not initial_animation:
            raise BearException('Initial animation ID should be provided')
        if initial_animation not in animations:
            raise BearException('Incorrect initial animation ID')
        super().__init__(*animations[initial_animation].frames[0])
        self.animations = animations
        self.current_animation = initial_animation
        self.running_index = 0
        self.have_waited = 0
        self.emit_ecs = emit_ecs
        self.cycle = cycle
        self.am_running = True

    def on_event(self, event):
        # When self.am_running is False, this widget does not respond to any
        # events and acts like a regular passive Widget
        if self.am_running:
            if event.event_type == 'tick':
                self.have_waited += event.event_value
                if self.have_waited >= self.animation.frame_time:
                    self.running_index += 1
                    if self.running_index >= len(self.animation):
                        if self.cycle:
                            self.running_index = 0
                        else:
                            self.am_running = False
                    self.chars = self.animation.frames[self.running_index][0]
                    self.colors = self.animation.frames[self.running_index][1]
                    self.have_waited = 0
                    if self.emit_ecs:
                        return BearEvent(event_type='ecs_update')
            elif self.parent is self.terminal and event.event_type == 'service'\
                    and event.event_value == 'tick_over':
                # This widget is connected to the terminal directly and must
                # update itself without a layout
                self.terminal.update_widget(self)

    @property
    def animation(self):
        return self.animations[self.current_animation]

    def set_animation(self, anim_id, cycle=False):
        """
        Set the next animation to be played.
        :param anim_id: Animation ID. Should be present in self.animations
        :param cycle: Whether to cycle the animation. Default False.
        :return:
        """
        if anim_id not in self.animations:
            raise BearException('Incorrect animation ID')
        self.current_animation = anim_id
        self.cycle = cycle
        self.am_running = True
        
    def __repr__(self):
        d = {'class': self.__class__.__name__,
             'animations': {x: repr(self.animations[x])
                            for x in self.animations},
             # Start from whichever animation was displayed during saving
             'initial_animation': self.current_animation,
             'emit_ecs': self.emit_ecs,
             'cycle': self.cycle}
        return dumps(d)


# Functional widgets. Please note that these include no decoration, BG, frame or
# anything else. Ie Label is just a chunk of text on the screen, FPSCounter and
# MousePosWidget are just the numbers that change. For the more complex visuals,
# embed these into a Layout with a preferred BG

class Label(Widget):
    """
    A widget that displays text.
    Accepts only a single string, whether single- or multiline.
    Does not (yet) support complex text markup used by bearlibterminal
    :param text: string to be displayed
    :param just: horizontal text justification, one of 'left', 'right'
    or 'center'. Default 'left'.
    :param color: bearlibterminal-compatible color. Default 'white'
    :param width: text area width. Defaults to the length of the longest
    substring in `text`.
    :param height: text area height. Defaults to the line count in `text`

    Label's text can be edited at any time by setting label.text property. Note
    that it overwrites any changes to `self.chars` and `self.colors` made after
    setting `self.text` the last time.
    
    Unlike text, Label's height and width cannot be changed. Set its height and
    width to accomodate all possible inputs during Label creation.
    """
    
    def __init__(self, text, chars=None, colors=None,
                 just='left', color='white', width=None, height=None):
        # If chars and colors are not provided, generate them. If they are,
        # typically from JSON dump, no checks are performed. Thus, in theory
        # it's possible to break this by providing overly big text and changing
        # adjustment.
        if not chars:
            chars = Label._generate_chars(text, width, height, just)
        if not colors:
            colors = copy_shape(chars, color)
        super().__init__(chars, colors)
        self.color = color
        # Bypassing setter, because I need to actually create fields
        self._just = just
        self._text = text
    
    @staticmethod
    def _generate_chars(text, width, height, just):
        """
        Internal method that generates a justified char list for the Label
        :param text:
        :param just:
        :return:
        """
        
        def justify(line, width, just_type):
            if len(line) < width:
                if just_type == 'left':
                    return line.ljust(width)
                elif just_type == 'right':
                    return line.rjust(width)
                elif just_type == 'center':
                    lj = width - int((width - len(line)) / 2)
                    return line.ljust(lj).rjust(width)
                else:
                    raise BearException(
                        'Justification should be \'left\', \'right\' or \'center\'')
            else:
                return line
        
        lines = text.split('\n')
        if not width:
            width = max(len(x) for x in lines)
        r = [list(justify(x, width, just)) for x in lines]
        if height and len(r) < height:
            for x in range(height - len(r)):
                r.append([' ' for j in range(len(r[0]))])
        return r

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        # TODO: check text size in Label
        # maybe raise something sensible when `text` is set for something longer
        # than fits in `chars`.
        if not self._text:
            self._text = value
        chars = copy_shape(self.chars, ' ')
        self.chars = blit(chars, self._generate_chars(value,
                                                      len(self.chars[0]),
                                                      len(self.chars),
                                                      self.just),
                          0, 0)
        self._text = value
        # MousePosWidgets (a child of Label) may have self.terminal set
        # despite not being connected to the terminal directly
        if self.terminal and self in self.terminal._widget_pointers:
            self.terminal.update_widget(self)

    @property
    def just(self):
        return self._just
    
    @just.setter
    def just(self, value):
        self.chars = Label._generate_chars(self.text, len(self.chars[0]),
                                           len(self.chars), just=value)
        if self.terminal:
            self.terminal.update_widget(self)
            
    def __repr__(self):
        d = loads(super().__repr__())
        d['text'] = self.text
        d['just'] = self.just
        d['color'] = self.color
            
            
class InputField(Label):
    """
    A single-line field for keyboard input.
    
    The length of the input line is limited by the InputField size. When the
    input is finished (by pressing ENTER), InputField emits a
    `BearEvent(event_type='text_input', event_value=(field.name, field.text))`
    Since BLT has no support for system keyboard layouts, only supports QWERTY
    Latin. This also applies to non-letter symbols: for example, comma and
    period are considered to be different keys even in Russian layout, where
    they are on the same physical key.
    """
    charcodes = {'SPACE': ' ', 'MINUS': '-', 'EQUALS': '=',
                 'LBRACKET': '[', 'RBRACKET': ']', 'BACKSLASH': '\\',
                 'SEMICOLON': ';', 'APOSTROPHE': '\'', 'GRAVE': '`',
                 'COMMA': ',', 'PERIOD': '.', 'SLASH': '/',
                 'KP_DIVIDE': '/', 'KP_MULTIPLY': '*', 'KP_MINUS': '-',
                 'KP_PLUS': '+', 'KP_1': '1', 'KP_2': '2', 'KP_3': '3',
                 'KP_4': '4', 'KP_5': '5', 'KP_6': 6, 'KP_7': 7,
                 'KP_8': 8, 'KP_9': 9, 'KP_0': 0, 'KP_PERIOD': '.'
                 }
    
    # Charcodes for non-letter characters used via Shift button
    shift_charcodes = {'MINUS': '_', 'EQUALS': '+', 'LBRACKET': '{',
                       'RBRACKET': '}', 'BACKSLASH': '|', 'SEMICOLON': ':',
                       'APOSTROPHE': '\"', 'GRAVE': '~', 'COMMA': '<',
                       'PERIOD': '>', 'SLASH': '?', '1': '!', '2': '@',
                       '3': '#', '4': '$', '5': '%', '6': '^', '7': '&',
                       '8': '*', '9': '(', '0': ')'}
    
    def __init__(self, name='Input field', accept_input=True, finishing=False,
                 **kwargs):
        if 'width' not in kwargs:
            raise BearException('InputField cannot be created without ' +
                                'either `width` or default text')
        super().__init__('', **kwargs)
        # The name will be used when the input is finished
        self.name = name
        self.shift_pressed = False
        # Set to True to return 'text_input' and stop accepting
        self.finishing = finishing
        self.accept_input = accept_input
        
    def on_event(self, event):
        #TODO Reactivate InputField on mouse click, if inactive
        # Requires it to have terminal for state.
        if self.finishing:
            # If finishing, the event will be ignored
            return BearEvent(event_type='text_input',
                             event_value=(self.name, self.text))
        if self.accept_input and event.event_type == 'key_down':
            # Stripping 'TK_' part
            symbol = event.event_value[3:]
            if symbol == 'BACKSPACE':
                self.text = self.text[:-1]
            elif symbol == 'SHIFT':
                self.shift_pressed = True
            # TK_ENTER is presumed to be the end of input
            elif symbol == 'ENTER':
                self.accept_input = False
                # returns immediately, unlike self.finish which sets it to
                # return on the *next* event
                return BearEvent(event_type='text_input',
                                 event_value=(self.name, self.text))
            elif len(self.text) < len(self.chars[0]):
                self.text += self._get_char(symbol)
            if self.terminal:
                self.terminal.update_widget(self)
        elif event.event_type == 'key_up':
            if event.event_value == 'TK_SHIFT':
                self.shift_pressed = False

    def finish(self):
        """
        Finish accepting the input and emit the 'text_input' event at the next
        opportunity.
        
        The event will not be actually emitted until the next event is passed
        to on_event.
        :return:
        """
        self.accept_input = False
        self.finishing = True
        
    def _get_char(self, symbol):
        """
        Return the char corresponding to a TK_* code.
        
        Considers the shift state
        :param symbol:
        :return:
        """
        if len(symbol) == 1:
            if self.shift_pressed:
                if symbol in '1234567890':
                    return self.shift_charcodes[symbol]
                else:
                    return symbol
            else:
                return symbol.lower()
        elif symbol in self.charcodes:
            if self.shift_pressed and symbol in self.shift_charcodes:
                return self.shift_charcodes[symbol]
            else:
                return self.charcodes[symbol]
        else:
            return ''
        
    def __repr__(self):
        d = loads(super().__repr__())
        d['name'] = self.name
        d['finishing'] = self.finishing
        d['accept_input'] = self.accept_input
        return dumps(d)


class FPSCounter(Label):
    """
    A simple widget that measures FPS.
    Actually just prints 1/(average runtime over the last 100 ticks in seconds),
    so it takes 100 ticks to get an accurate reading. Not relevant except on the
    first several seconds of the program run or after FPS has changed, but if it
    seems like the game takes a second or two to reach the target FPS -- it just
    seems that way.
    """
    def __init__(self, **kwargs):
        self.samples_deque = deque(maxlen=100)
        super().__init__('030', **kwargs)
    
    def _update_self(self):
        fps = str(round(len(self.samples_deque) /
                            sum(self.samples_deque)))
        fps = fps.rjust(3, '0')
        self.text = fps
    
    def on_event(self, event):
        # Update FPS estimate
        if event.event_type == 'tick':
            self.samples_deque.append(event.event_value)
            self._update_self()
            if self.parent is self.terminal:
                self.terminal.update_widget(self, refresh=True)
                
    def __repr__(self):
        raise BearException('FPSCounter does not support __repr__ serialization')
        # This should only be used **OUTSIDE** the ECS system.
        # Some debug screen or something


class MousePosWidget(Label):
    """
    A simple widget akin to FPSCounter that listens to TK_MOUSE_MOVE events.
    
    In order to work, it needs `self.terminal` to be set to the current
    terminal, which means it should either be added to the terminal directly
    (without any Layouts) or it should be set manually before MousePosWidget
    gets its first `tick` event.
    """
    
    def __init__(self, **kwargs):
        super().__init__(text='000x000', **kwargs)
        
    def on_event(self, event):
        if event.event_type == 'misc_input' and \
                     event.event_value == 'TK_MOUSE_MOVE':
            self.text = self.get_mouse_line()
        if isinstance(self.parent, BearTerminal):
            self.terminal.update_widget(self)

    def get_mouse_line(self):
        if not self.terminal:
            raise BearException('MousePosWidget is not connected to a terminal')
        x = str(self.terminal.check_state('TK_MOUSE_X')).rjust(3, '0')
        y = str(self.terminal.check_state('TK_MOUSE_Y')).rjust(3, '0')
        return x + 'x' + y

    def __repr__(self):
        raise BearException('MousePosWidget does not support __repr__ serialization')
    
# Listeners
class Listener:
    """
    A base class for the things that need to interact with the queue (and maybe
    the terminal), but aren't widgets.
    """
    def __init__(self):
        self.terminal = None
    
    def on_event(self, event):
        raise NotImplementedError('Listener base class is doing nothing')
    
    def register_terminal(self, terminal):
        """
        Register a terminal with which this listener will interact
        :param terminal: A BearTerminal instance
        :return:
        """
        if not isinstance(terminal, BearTerminal):
            raise TypeError('Only BearTerminal instances registered by Listener')
        self.terminal = terminal
        
    
class ClosingListener(Listener):
    """
    The listener that waits for TK_CLOSE input event (Alt-F4 or closing window)
    and sends the shutdown service event to the queue. All widgets are expected
    to listen to it and immediately save their data or whatever they need to do.
    On the next tick ClosingListener closes the terminal and queue altogether.
    """
    def __init__(self):
        super().__init__()
        self.countdown = 2
        self.counting = False
        
    def on_event(self, event):
        if event.event_type == 'misc_input' and event.event_value == 'TK_CLOSE':
            self.counting = True
            return BearEvent(event_type='service', event_value='shutdown_ready')
        if event.event_type == 'tick':
            if self.counting:
                self.countdown -= 1
                if self.countdown == 0:
                    return BearEvent(event_type='service',
                                     event_value='shutdown')


class LoggingListener(Listener):
    """
    A listener that logs the events it accepts
    """
    def __init__(self, handle):
        super().__init__()
        if not hasattr(handle, 'write'):
            raise BearException('The LoggingListener needs a writable object')
        self.handle = handle
        
    def on_event(self, event):
        self.handle.write('{0}: type {1}, '.format(str(time()), event.event_type) +
                          'value {}\n'.format(event.event_value))



"""
An object-oriented bearlibterminal wrapper with the support for complex ASCII
art and widget-like behaviour.
"""

from bearlibterminal import terminal
from bear_hug.bear_utilities import BearException,\
    BearLoopException
from bear_hug.event import BearEvent

import inspect
import os
import time
from copy import copy


class BearTerminal:
    """
    A main terminal class.
    Accepts bearlibterminal library configuration options as kwargs to __init__.
    Currently only a library settings aresupported and there is no support for
    changing them on the fly, but that's subject to fixing.
    """
    # TODO: wrap bearlibterminal parameters in @property
    # TODO: change bearlibterminal parameters on the fly
    # kwargs to init are passed to bearlibterminal.terminal.set()
    # Currently only library settings are supported
    accepted_kwargs = {'encoding': 'terminal', 'size': 'window',
                       'cellsize': 'window', 'title': 'window', 'icon': 'window',
                       'resizeable': 'window', 'fullscreen': 'window',
                       
                       'filter': 'input', 'precise-mouse': 'input',
                       'mouse-cursor': 'input', 'cursor-symbol': 'input',
                       'cursor-blink-rate': 'input', 'alt-functions': 'input',
    
                       'postformatting': 'output', 'vsync': 'output',
                       'tab-width': 'output',
                       
                       'file': 'log', 'level': 'log', 'mode': 'log'
                       }
    
    # Bearlibterminal's input codes and terminal state codes are converted to
    # constants before emitting events, so that downstream widgets could process
    # them without asking the terminal to explain what the fuck 0x50 is.
    
    # These are codes for keys and mouse buttons going down
    down_codes = {0x04: 'TK_A', 0x05: 'TK_B', 0x06: 'TK_C', 0x07: 'TK_D',
          0x08: 'TK_E', 0x09: 'TK_F', 0x0A: 'TK_G', 0x0B: 'TK_H', 0x0C: 'TK_I',
          0x0D: 'TK_J', 0x0E: 'TK_K', 0x0F: 'TK_L', 0x10: 'TK_M', 0x11: 'TK_N',
          0x12: 'TK_O', 0x13: 'TK_P', 0x14: 'TK_Q', 0x15: 'TK_R', 0x16: 'TK_S',
          0x17: 'TK_T', 0x18: 'TK_U', 0x19: 'TK_V', 0x1A: 'TK_W', 0x1B: 'TK_X',
          0x1C: 'TK_Y', 0x1D: 'TK_Z',
          0x1E: 'TK_1', 0x1F: 'TK_2', 0x20: 'TK_3',
          0x21: 'TK_4', 0x22: 'TK_5', 0x23: 'TK_6', 0x24: 'TK_7', 0x25: 'TK_8',
          0x26: 'TK_9', 0x27: 'TK_0',
          0x28: 'TK_ENTER', 0x29: 'TK_ESCAPE', 0x2A: 'TK_BACKSPACE',
          0x2B: 'TK_TAB', 0x2C: 'TK_SPACE', 0x2D: 'TK_MINUS', 0x2E: 'TK_EQUALS',
          0x2F: 'TK_LBRACKET', 0x30: 'TK_RBRACKET', 0x31: 'TK_BACKSLASH',
          0x33: 'TK_SEMICOLON', 0x34: 'TK_APOSTROPHE', 0x35: 'TK_GRAVE',
          0x36: 'TK_COMMA', 0x37: 'TK_PERIOD', 0x38: 'TK_SLASH',
          0x3A: 'TK_F1', 0x3B: 'TK_F2', 0x3C: 'TK_F3', 0x3D: 'TK_F4',
          0x3E: 'TK_F5', 0x3F: 'TK_F6', 0x40: 'TK_F7', 0x41: 'TK_F8', 0x42: 'TK_F9',
          0x43: 'TK_F10', 0x44: 'TK_F11', 0x45: 'TK_F12',
          0x48: 'TK_PAUSE', 0x49: 'TK_INSERT', 0x4A: 'TK_HOME',
          0x4B: 'TK_PAGEUP', 0x4C: 'TK_DELETE', 0x4D: 'TK_END', 0x4E: 'TK_PAGEDOWN',
          0x4F: 'TK_RIGHT', 0x50: 'TK_LEFT', 0x51: 'TK_DOWN', 0x52: 'TK_UP',
          0x54: 'TK_KP_DIVIDE', 0x55: 'TK_KP_MULTIPLY',
          0x56: 'TK_KP_MINUS', 0x57: 'TK_KP_PLUS', 0x58: 'TK_KP_ENTER',
          0x59: 'TK_KP_1', 0x5A: 'TK_KP_2', 0x5B: 'TK_KP_3', 0x5C: 'TK_KP_4',
          0x5D: 'TK_KP_5', 0x5E: 'TK_KP_6', 0x5F: 'TK_KP_7', 0x60: 'TK_KP_8',
          0x61: 'TK_KP_9', 0x62: 'TK_KP_0', 0x63: 'TK_KP_PERIOD',
          0x70: 'TK_SHIFT', 0x71: 'TK_CONTROL', 0x72: 'TK_ALT',
          0x80: 'TK_MOUSE_LEFT', 0x81: 'TK_MOUSE_RIGHT',
          0x82: 'TK_MOUSE_MIDDLE'}
    
    # The same buttons going up.
    # BLT OR's key code with TK_KEY_RELEASED, which is not reversible except by
    # bruteforcing all the keys. Thus, a second dict.
    up_codes = {260: 'TK_A', 261: 'TK_B', 262: 'TK_C', 263: 'TK_D', 264: 'TK_E',
                265: 'TK_F', 266: 'TK_G', 267: 'TK_H', 268: 'TK_I', 269: 'TK_J',
                270: 'TK_K', 271: 'TK_L', 272: 'TK_M', 273: 'TK_N', 274: 'TK_O',
                275: 'TK_P', 276: 'TK_Q', 277: 'TK_R', 278: 'TK_S', 279: 'TK_T',
                280: 'TK_U', 281: 'TK_V', 282: 'TK_W', 283: 'TK_X', 284: 'TK_Y',
                285: 'TK_Z', 286: 'TK_1', 287: 'TK_2', 288: 'TK_3', 289: 'TK_4',
                290: 'TK_5', 291: 'TK_6', 292: 'TK_7', 293: 'TK_8', 294: 'TK_9',
                295: 'TK_0', 296: 'TK_ENTER', 297: 'TK_ESCAPE',
                298: 'TK_BACKSPACE', 299: 'TK_TAB', 300: 'TK_SPACE',
                301: 'TK_MINUS', 302: 'TK_EQUALS', 303: 'TK_LBRACKET',
                304: 'TK_RBRACKET', 305: 'TK_BACKSLASH', 307: 'TK_SEMICOLON',
                308: 'TK_APOSTROPHE', 309: 'TK_GRAVE', 310: 'TK_COMMA',
                311: 'TK_PERIOD', 312: 'TK_SLASH', 314: 'TK_F1', 315: 'TK_F2',
                316: 'TK_F3', 317: 'TK_F4', 318: 'TK_F5', 319: 'TK_F6',
                320: 'TK_F7', 321: 'TK_F8', 322: 'TK_F9', 323: 'TK_F10',
                324: 'TK_F11', 325: 'TK_F12', 328: 'TK_PAUSE', 329: 'TK_INSERT',
                30: 'TK_HOME', 331: 'TK_PAGEUP', 332: 'TK_DELETE',
                333: 'TK_END', 334: 'TK_PAGEDOWN', 335: 'TK_RIGHT',
                336: 'TK_LEFT', 337: 'TK_DOWN', 338: 'TK_UP',
                340: 'TK_KP_DIVIDE', 341: 'TK_KP_MULTIPLY', 342: 'TK_KP_MINUS',
                343: 'TK_KP_PLUS', 344: 'TK_KP_ENTER', 345: 'TK_KP_1',
                346: 'TK_KP_2', 347: 'TK_KP_3', 348: 'TK_KP_4', 349: 'TK_KP_5',
                350: 'TK_KP_6', 351: 'TK_KP_7', 352: 'TK_KP_8', 353: 'TK_KP_9',
                354: 'TK_KP_0', 355: 'TK_KP_PERIOD', 368: 'TK_SHIFT',
                369: 'TK_CONTROL', 370: 'TK_ALT', 384: 'TK_MOUSE_LEFT',
                385: 'TK_MOUSE_RIGHT', 386: 'TK_MOUSE_MIDDLE'}
    
    # This is misc input and state codes
    misc_input = {0x83: 'TK_MOUSE_X1', 0x84: 'TK_MOUSE_X2',
    0x85: 'TK_MOUSE_MOVE', 0x86: 'TK_MOUSE_SCROLL',
                   0x87: 'TK_MOUSE_X', 0x88: 'TK_MOUSE_Y',
    0x89: 'TK_MOUSE_PIXEL_X', 0x8A: 'TK_MOUSE_PIXEL_Y', 0x8B: 'TK_MOUSE_WHEEL',
    0x8C: 'TK_MOUSE_CLICKS',
                   0xC0: 'TK_WIDTH', 0xC1: 'TK_HEIGHT', 0xC2: 'TK_CELL_WIDTH',
    0xC3: 'TK_CELL_HEIGHT', 0xC4: 'TK_COLOR', 0xC5: 'TK_BKCOLOR',
    0xC6: 'TK_LAYER', 0xC7: 'TK_COMPOSITION', 0xC8: 'TK_CHAR',
    0xC9: 'TK_WCHAR', 0xCA: 'TK_EVENT', 0xCB: 'TK_FULLSCREEN',
                   0xE0: 'TK_CLOSE', 0xE1: 'TK_RESIZED'}
    
    # This is the name-to-number mapping, as in bearlibterminal/terminal.py
    # The purpose of this dict is, again, to let bear_hug users work with strs
    # and avoid thinking about constants
    state_constants = {'TK_BACKSLASH': 49, 'TK_KP_1': 89,
        'TK_KEY_RELEASED': 256, 'TK_Q': 20, 'TK_4': 33, 'TK_N': 17,
        'TK_ALIGN_TOP': 4, 'TK_RESIZED': 225, 'TK_ALIGN_RIGHT': 2,
        'TK_H': 11, 'TK_MOUSE_MIDDLE': 130, 'TK_CELL_WIDTH': 194, 'TK_3': 32,
        'TK_F10': 67, 'TK_RIGHT': 79, 'TK_ESCAPE': 41, 'TK_KP_6': 94,
        'TK_COMMA': 54, 'TK_GRAVE': 53, 'TK_PERIOD': 55, 'TK_MOUSE_CLICKS': 140,
        'TK_KP_3': 91, 'TK_MOUSE_MOVE': 133, 'TK_KP_2': 90,
        'TK_INPUT_CANCELLED': -1, 'TK_EVENT': 202, 'TK_KP_0': 98,
        'TK_DELETE': 76, 'TK_DOWN': 81, 'TK_CLOSE': 224, 'TK_R': 21,
        'TK_F9': 66, 'TK_M': 16, 'TK_ALIGN_MIDDLE': 12, 'TK_1': 30, 'TK_5': 34,
        'TK_WIDTH': 192, 'TK_F': 9, 'TK_G': 10, 'TK_SLASH': 56,
        'TK_PAGEDOWN': 78, 'TK_MOUSE_X2': 132, 'TK_KP_PLUS': 87,
        'TK_MOUSE_X1': 131, 'TK_HOME': 74, 'TK_W': 26, 'TK_F1': 58,
        'TK_MOUSE_X': 135, 'TK_ALIGN_LEFT': 1, 'TK_ALIGN_BOTTOM': 8,
        'TK_FULLSCREEN': 203, 'TK_C': 6, 'TK_CONTROL': 113,
        'TK_MOUSE_PIXEL_Y': 138, 'TK_COLOR': 196, 'TK_PAGEUP': 75,
        'TK_RBRACKET': 48, 'TK_F11': 68, 'TK_MOUSE_SCROLL': 134,
        'TK_KP_MULTIPLY': 85, 'TK_F8': 65, 'TK_MINUS': 45, 'TK_ENTER': 40,
        'TK_ALT': 114, 'TK_O': 18, 'TK_Z': 29, 'TK_COMPOSITION': 199,
        'TK_F2': 59, 'TK_T': 23, 'TK_EQUALS': 46, 'TK_ALIGN_CENTER': 3,
        'TK_Y': 28, 'TK_BACKSPACE': 42, 'TK_SHIFT': 112, 'TK_S': 22,
        'TK_SPACE': 44, 'TK_MOUSE_RIGHT': 129, 'TK_KP_5': 93, 'TK_KP_ENTER': 88,
        'TK_9': 38, 'TK_MOUSE_Y': 136, 'TK_6': 35, 'TK_UP': 82,
        'TK_BKCOLOR': 197, 'TK_2': 31, 'TK_KP_4': 92, 'TK_MOUSE_LEFT': 128,
        'TK_F7': 64, 'TK_HEIGHT': 193, 'TK_X': 27, 'TK_V': 25, 'TK_P': 19,
        'TK_D': 7, 'TK_KP_9': 97, 'TK_PAUSE': 72, 'TK_F6': 63, 'TK_WCHAR': 201,
        'TK_U': 24, 'TK_KP_7': 95, 'TK_APOSTROPHE': 52, 'TK_F5': 62,
        'TK_LBRACKET': 47, 'TK_7': 36, 'TK_8': 37, 'TK_TAB': 43, 'TK_0': 39,
        'TK_MOUSE_PIXEL_X': 137, 'TK_CHAR': 200, 'TK_KP_DIVIDE': 84,
        'TK_LAYER': 198, 'TK_KP_8': 96, 'TK_B': 5, 'TK_ALIGN_DEFAULT': 0,
        'TK_K': 14, 'TK_KP_MINUS': 86, 'TK_LEFT': 80, 'TK_CELL_HEIGHT': 195,
        'TK_SEMICOLON': 51, 'TK_INSERT': 73, 'TK_END': 77, 'TK_F12': 69,
        'TK_MOUSE_WHEEL': 139, 'TK_F3': 60, 'TK_L': 15, 'TK_KP_PERIOD': 99,
        'TK_J': 13, 'TK_F4': 61}
    
    def __init__(self, font_path='../demo_assets/cp437_12x12.png',
                 *args, **kwargs):
        if kwargs:
            if any(x not in self.accepted_kwargs for x in kwargs.keys()):
                raise BearException('Only bearlibterminal library settings '
                                    +' accepted as kwargs for BearTerminal')
            self.outstring = ';'.join('{}.{}={}'.format(self.accepted_kwargs[x],
                                                        x, str(kwargs[x]))
                                 for x in kwargs)+';'
        else:
            self.outstring = None
        self.widget_locations = {}
        #  This will be one list of drawable pointers per layer. Lists are
        #  not actually allocated until at least one Widget is added to layer
        #  Lists are created when adding the first Widget and are never
        #  destroyed or resized.
        self._widget_pointers = [None for x in range(256)]
        self.default_color = 'white'
        # TODO: make font_path system independent
        self.font_path = font_path

    #  Methods that replicate or wrap around blt's functions

    def start(self):
        """
        Open a terminal and place it on the screen.
        Applies library settings that were set in `__init__()`
        :return:
        """
        terminal.open()
        terminal.set(
            'font: {}, size=12x12, codepage=437'.format(self.font_path))
        if self.outstring:
            terminal.set(self.outstring)
        self.refresh()
        
    def clear(self):
        """
        Remove all widgets from this terminal, but do not close it
        :return:
        """
        drawables = copy(self.widget_locations)
        for drawable in drawables:
            self.remove_widget(drawable, refresh=False)
        self.refresh()

    def refresh(self):
        """
        Refresh a terminal.
        Places whatever changes were made by `*_widget` methods on the screen.
        :return:
        """
        terminal.refresh()

    def close(self):
        """
        Close a terminal.
        Does not destroy Widget objects or call any other cleanup routine.
        :return:
        """
        terminal.close()

    #  Drawing and removing stuff

    def add_widget(self, widget,
                   pos=(0, 0), layer=0, refresh=False):
        """
        Add a widget to the terminal and set `widget.terminal` to `self`.
        No two widgets are allowed to overlap within a layer and no widget can
        be added twice.
        :param widget: a Widget instance
        :param pos: top left corner of the widget
        :param layer: layer to place the widget on
        :param refresh: whether to refresh terminal after adding the widget.
        If this is False, the widget will not be actually shown until the next
        `terminal.refresh()` call
        :return:
        """
        if widget in self.widget_locations.keys():
            raise BearException('Cannot add the same widget twice')
        for y in range(widget.height):
            for x in range(widget.width):
                if self._widget_pointers[layer] and \
                        self._widget_pointers[layer][pos[0] + x][pos[1] + y]:
                    raise BearException('Widgets cannot collide within a layer')
        widget.terminal = self
        widget.parent = self
        self.widget_locations[widget] = WidgetLocation(pos=pos, layer=layer)
        terminal.layer(layer)
        if not self._widget_pointers[layer]:
            size = terminal.get('window.size')
            width, height = (int(x) for x in size.split('x'))
            self._widget_pointers[layer] = [[None for y in range(height)]
                                            for x in range(width)]
        self.update_widget(widget, refresh)
    
    def remove_widget(self, widget, refresh=False):
        """
        Remove widget from the terminal
        :param widget: A widget to be removed
        :param refresh: whether to refresh the terminal after removing a widget.
        If this is False, the widget will be visible until the next
        `terminal.refresh()` call
        :return:
        """
        corner = self.widget_locations[widget].pos
        terminal.layer(self.widget_locations[widget].layer)
        terminal.clear_area(*corner, widget.width, widget.height)
        for y in range(len(widget.chars)):
            for x in range(len(widget.chars[0])):
                self._widget_pointers[self.widget_locations[widget].layer]\
                    [corner[0] + x][corner[1] + y] = None
        if refresh:
            self.refresh()
        del(self.widget_locations[widget])
        widget.terminal = None
        widget.parent = None
        
    def move_widget(self, widget, pos, refresh=False):
        """
        Move widget to a new position.
        Widgets can only be moved within the layer.
        :param widget:
        :param pos:
        :return:
        """
        layer = self.widget_locations[widget].layer
        self.remove_widget(widget)
        self.add_widget(widget, pos=pos, layer=layer)
        if refresh:
            self.refresh()

    def update_widget(self, widget, refresh=False):
        """
        Actually place widget chars on screen.
        :param widget:
        :return:
        """
        if widget not in self.widget_locations:
            raise BearException('Cannot update non-added Widgets')
        pos = self.widget_locations[widget].pos
        layer = self.widget_locations[widget].layer
        terminal.layer(layer)
        terminal.clear_area(*self.widget_locations[widget].pos, widget.width, widget.height)
        running_color = self.default_color
        for y in range(widget.height):
            for x in range(widget.width):
                # Widget can have None as color for its empty cells
                if widget.colors[y][x] and widget.colors[y][x] != running_color:
                    running_color = widget.colors[y][x]
                    terminal.color(running_color)
                terminal.put(pos[0] + x, pos[1] + y, widget.chars[y][x])
                self._widget_pointers[layer][pos[0] + x][pos[1] + y] = widget
        if running_color != self.default_color:
            terminal.color(self.default_color)
        if refresh:
            self.refresh()
    
    #  Getting terminal info

    def get_widget_by_pos(self, pos, layer=None):
        """
        Returns the widget currently placed at the given position.
        If layer is set, checks only that layer. Otherwise returns the widget
        at the highest layer.
        :param pos: 
        :param layer: 
        :return: 
        """
        if layer:
            return self._widget_pointers[layer][pos[0]][pos[1]]
        else:
            for layer_list in reversed(self._widget_pointers):
                if layer_list and layer_list[pos[0]][pos[1]]:
                    return layer_list[pos[0]][pos[1]]
            return None
        
    # Input
    def check_input(self):
        """
        Check if terminal has input. If so, yield corresponding `BearEvent`(s)
        This method returns an iterator because it's possible there would be
        more than one event in a single tick, eg when two buttons are pressed
        simultaneously.
        :param self:
        :return:
        """
        while terminal.has_input():
            # Process the input event
            in_event = terminal.read()
            if in_event in self.misc_input:
                yield BearEvent('misc_input', self.misc_input[in_event])
            elif in_event in self.down_codes:
                yield BearEvent('key_down', self.down_codes[in_event])
            elif in_event in self.up_codes:
                yield BearEvent('key_up', self.up_codes[in_event])
            else:
                raise BearException('Unknown input code {}'.format(in_event))
    
    def check_state(self, query):
        """
        A wrapper around BLT `state` function
        Accepts any of the TK_* strings and returns whatever terminal.state has
        to say about it.
        :param query:
        :return:
        """
        return terminal.state(self.state_constants[query])


#  A loop

class BearLoop:
    """
    A loop that passes event around every 1/fps seconds.
    FPS is only set on loop creation, which will probably be fixed later.
    Every tick, the loop emits two events. In the beginning of the tick it's
    'tick' event whose value is time since the last similar event (in seconds).
    In the end of the tick it's 'service' event with the value 'tick_over',
    meant to allow widgets to update themselves to reflect whatever
    changed in the last event. The first event is guaranteed to be emitted
    before any other events related to this tick; the former is issued only
    after the entire queue is processed, but the responses to it (if any) will
    be dispatched after itself and before the next 'tick' event.
    Generally, it's safer not to emit events in response to 'tick_over' event,
    because whatever happens in response to *them* will only be drawn on the
    next tick.
    """
    def __init__(self, terminal, queue, fps=30):
        # Assumes terminal to be running
        self.terminal = terminal
        self.queue = queue
        # The loop listens for service events so that it knows when to shutdown
        self.queue.register_listener(self, 'service')
        self.frame_time = 1/fps
        self.stopped = False
        self.last_time = 0
        
    def run(self):
        """
        Start a loop.
        It would run indefinitely until stopped with `self.stop`
        :return:
        """
        # An imaginary "zeroth" tick to give the first tick correct timing
        self.last_time = time.time() - self.frame_time
        while not self.stopped:
            # All actual processes happen here
            # Sends time since last tick *started*
            t = time.time() - self.last_time
            self.last_time = time.time()
            self.run_iteration(t)
            sleep_time = self.frame_time - time.time() + self.last_time
            if sleep_time > 0.05*self.frame_time:
                # If frame was finished early, wait for it
                # But only if there is enough spare time to make it worthwhile.
                # Otherwise, on a laggy system sleep_time may be positive when
                # the `if` check runs, but negative by the time `sleep` is
                # called, causing a crash.
                time.sleep(self.frame_time - time.time() + self.last_time)
        # When the loop stops, it closes the terminal. Everyone is expected to
        # have caught the shutdown service event
        self.terminal.close()
               
    def stop(self):
        """
        Stop the loop.
        It would quit after finishing the current iteration
        :return:
        """
        self.stopped = True
    
    def run_iteration(self, time_since_last_tick):
        # Get input events, if any
        for event in self.terminal.check_input():
            self.queue.add_event(event)
        self.queue.add_event(BearEvent(event_type='tick',
                                       event_value=time_since_last_tick))
        self.queue.dispatch_events()
        # Sending "Tick over" event, reminding widgets to update themselves
        self.queue.add_event(BearEvent(event_type='service',
                                       event_value='tick_over'))
        self.queue.dispatch_events()
        self.terminal.refresh()
        
    def on_event(self, event):
        if event.event_value == 'shutdown':
            self.stopped = True
            
    @property
    def fps(self):
        return round(1/self.frame_time)
    
    @fps.setter
    def fps(self, value):
        if not isinstance(value, int):
            raise BearLoopException('Only int acceptable as FPS')
        self.frame_time = 1/value
        
    
#  Misc classes

class WidgetLocation:
    """
    Data class with position and layer of a Widget
    """
    def __init__(self, pos, layer):
        self.pos = pos
        self.layer = layer
